-- Copyright 2010 Jo-Philipp Wich <jow@openwrt.org>
-- Copyright 2017 Dan Luedtke <mail@danrl.com>
-- Licensed to the public under the Apache License 2.0.

local fs = require "nixio.fs"
local ip = require "luci.ip"
local math = require "math"
local util = require "luci.util"
local tonumber, tostring, type, unpack, select = tonumber, tostring, type, unpack, select

module "luci.cbi.datatypes"

default_hints = {
	bool 		= "Only values: 1, 0, yes, no, on, off, true and false are accepted.",
	uinteger	= "Only positive integers are accepted.",
	integer		= "Negative and positive integers are accepted.",
	ufloat		= "Only positive float numbers are accepted. E.g. 1.32.",
	float		= "Negative and positive float numbers are accepted. E.g. 1.32.",
	ipaddr		= "IPv4 and IPv6 addresses are accepted. E.g. 192.168.1.1 .",
	ip4addr		= "IPv4 addresses are accepted. E.g. 192.168.1.1.",
	ip4prefix	= "IPv4 mask prefix. E.g. /24.",
	ip6addr		= "IPv6 addresses are accepted. E.g. ::0000:8a2e:0370:7334.",
	ip6prefix	= "IPv6 mask prefix. E.g. /128.",
	ipmask		= "IPv4 and IPv6 addresses with or without mask prefix are accepted. E.g 192.168.1.0/24.",
	ipmask4		= "IPv4 addresses with or without mask prefix are accepted. E.g 192.168.1.0/24.",
	ipmask6		= "IPv6 addresses with or without mask prefix are accepted. E.g ::00FF/64.",
	netmask		= "Netmasks are accepted E.g. 255.255.255.0.",
	ip6hostid	= "Allowed values: 'eui64', 'random', fixed value like '::1' or '::1:2'.",
	port		= "Values between 0 and 65535 are accepted.",
	portrange	= "Port ranges with values from 0 to 65535 are accepted. E.g. 232-254.",
	macaddr		= "Mac address of six groups of two hexadecimal digits are accepted. E.g. 01:23:45:67:89:AB.",
	hostname	= "Domain names are accepted. E.g. example.com .",
	host		= "Domain names or ip addresses accepted. E.g. 192.168.1.1 or example.com .",
	wpakey		= "A key between 8 and 64 symbols is accepted.",
	wepkey		= "A key of 5, 10, 13, 26 symbols is accepted.",
	hexstring	= "A hexadecimal string of symbols: a-f, A-F and 0-9 is accepted.",
	hex			= "A hexadecimal pair of symbols: a-f, A-F and 0-9 is accepted.",
	base64		= "A 64 symbol string made up of a-zA-Z0-9/+= characters is accepted.",
	string		= "Any string is accepted.",
	nospace		= "Any string without a space is accepted.",
	directory	= "A link to a directory is accepted. E.g. /tmp/example",
	file		= "A path to a file is accepted. E.g /tmp/file.sh",
	uciname		= "A string of a-Z, 0-9 and _ characters is accepted.",
	hostport	= "An IP address with a port is required E.g 192.168.1.1:80",
	timehhmmss	= "Time of format hh:mm:ss is accepted.",
	dateyyyymmdd= "A date of format yyyy:mm:dd is accepted. ",
	root_password= "A password of 8-32 characters, at least one uppercase letter, one lowercase letter and one number is accepted.",
	password_validate = "Any string without a space is accepted.",
	geofencing	= "Floating part of number must contain 6 digits E.g. 25.000000 .",
	protourl	= "A full url address is accepted. E.g. http://www.example.com .",
	phonedigit	= "A phone number containing 0-9\*+# characters is accepted. ",
	loglimit 	= "An time value with time interval(hour, minute ,second) is accepted. E.g. 10/minute",
	email		= "A valid email address is accepted. E.g. example@domain.com",
	range		= "Range of the value must be from %s to %s.",
	min			= "Minimum value is %s.",
	max			= "Maximum value is %s.",
	rangelength = "Length of the value must be from %s to %s.",
	minlength 	= "Minimum length of value is %s.",
	maxlength	= "Maximum length of value is %s.",
	url			= "Full url path allowed. E.g. example.com/example or 192.168.1.1/example."
}

default_lengths = {
	ip6addr		= "64",
	ipaddr		= "64",
	base64		= "64",
	hostname	= "64",
	host		= "64",
	email		= "64",
	wpakey		= "64",
	root_password="32",
	ip4addr 	= "18",
	ipmask4 	= "18",
	loglimit	= "18",
	macaddr		= "18",
	geofencing	= "12",
	dateyyyymmdd= "12",
	uciname		= "16"
}

local default_length = "16"

function set_length(matched, datatype)
	local maxlength = default_length
	if default_lengths[matched] then
		return default_lengths[matched]
	elseif matched == "maxlength" then
		maxlength = datatype:match("(%d+)")
	elseif matched == "rangelength" then
		_, maxlength = datatype:match("(%d+),(%d+)")
	end
	return maxlength
end

function and_or_parser(text, type)
	local value = ""
	local word = ""
	local open = 0
	local ret = ""
	local maxlength = tonumber(default_length)
	for ch in text:gmatch(".") do
		if ch == "(" then open = open + 1 end
		if ch == ")" then open = open - 1 end
		if ch == "," and open == 0 then
			ret, ml = default_hints:format_datatype(word)
			word = ""
			value = value .." ".. ret:sub(1,#ret-1) .. " "..type.." "
			mlnumber = tonumber(ml)
			if mlnumber > maxlength then
				maxlength = mlnumber
			end
		else
			word = word .. ch
		end
	end
	if #word > 0 then
		ret, ml = default_hints:format_datatype(word)
		value = value .." ".. ret:sub(1,#ret-1) .. " "..type.." "
		mlnumber = tonumber(ml)
		if mlnumber > maxlength then
			maxlength = mlnumber
		end
	end
	return value:sub(1, #value-(#type+1)) .. ".", tostring(maxlength)
end

function field_parser(type, datatype)
	local add_next, add, textend, first = false, false, false, true
	local num, parse_text = "", ""
	local length_val = ""
	local num1, num2 = ""
	if type == "length" then
		_, _, num1, num2 = datatype:find("([- ]?%d+),([- ]?%d+)")
		parse_text = parse_text.."Length of the value must be from "..num1.." to "..num2..". "
	end
	for ch in datatype:gmatch(".") do
		if ch == "]" and add_next == false then 
			add = false 
			textend = true 
			parse_text = parse_text .. ")."
		end
		if ch ~= "%" and add == true or add_next == true then
			parse_text = parse_text .. ch
			add_next = false
		elseif ch == "%" then 
			add_next = true
		end
		if ch == "[" and add_next == false then 
			add = true 
			if first then
				parse_text = parse_text .. "Allowed characters ("
				first = false
			end
		end
		if type == "field" then
			if textend then
				if tonumber(ch) ~= nil then
					num = num..ch
				end
			end
		end -- type == "field"
	end -- for ch in datatype:gmatch(".") do
	if num ~= "0" and num ~= "" then
		parse_text = parse_text .. " Minimum length is " .. num .."."
	end
	if type == "length" then
		return parse_text, num2
	else
		return parse_text
	end
end

function default_hints:format_datatype(datatype)
	local matched, num1, num2
	local maxlength = default_length
	local default = "Invalid value!"
	if not datatype or datatype == "" then 
		return default, maxlength
	end
	if datatype:find("%(") then
		matched = datatype:match("(%a+)%(")
		_, _, num1, num2 = datatype:find("([- ]?%d+),([- ]?%d+)")
		if not num1 then
			_, _, num1 = datatype:find("([- ]?%d+)")
		end
	else
		matched = datatype:gsub("%s+", "")
	end

	local inside = datatype:match("%((.+)%)")
	if matched then
		if matched == "neg" then
			return self:format_datatype(inside)
		elseif matched == "list" then
			err, ml = self:format_datatype(inside)
			return err .. " Allows the usage of a space separated multi-value.", ml
		elseif matched == "or" or matched == "and" then
			return and_or_parser(inside, matched)
		end
		if matched == "fieldvalidation" then
			return field_parser("field", datatype), maxlength
		end
		if matched == "lengthvalidation" then
			return field_parser("length", datatype)
		end
		maxlength = set_length(matched, datatype)
	end
	local hint = self[matched] or default
	if num1 or (num1 and num2) then
		return hint:format(num1, num2), maxlength
	else
		return hint, maxlength
	end
	return default, maxlength
end

_M['or'] = function(v, ...)
	local i
	for i = 1, select('#', ...), 2 do
		local f = select(i, ...)
		local a = select(i+1, ...)
		if type(f) ~= "function" then
			if f == v then
				return true
			end
			i = i - 1
		elseif f(v, unpack(a)) then
			return true
		end
	end
	return false
end

_M['and'] = function(v, ...)
	local i
	for i = 1, select('#', ...), 2 do
		local f = select(i, ...)
		local a = select(i+1, ...)
		if type(f) ~= "function" then
			if f ~= v then
				return false
			end
			i = i - 1
		elseif not f(v, unpack(a)) then
			return false
		end
	end
	return true
end

function neg(v, ...)
	return _M['or'](v:gsub("^%s*!%s*", ""), ...)
end

function list(v, subvalidator, subargs)
	if type(subvalidator) ~= "function" then
		return false
	end
	local token
	for token in v:gmatch("%S+") do
		if not subvalidator(token, unpack(subargs)) then
			return false
		end
	end
	return true
end

function bool(val)
	if val == "1" or val == "yes" or val == "on" or val == "true" then
		return true
	elseif val == "0" or val == "no" or val == "off" or val == "false" then
		return true
	elseif val == "" or val == nil then
		return true
	end

	return false
end

function uinteger(val)
	local n = tonumber(val)
	if n ~= nil and math.floor(n) == n and n >= 0 then
		return true
	end

	return false
end

function integer(val)
	local n = tonumber(val)
	if n ~= nil and math.floor(n) == n then
		return true
	end

	return false
end

function ufloat(val)
	local n = tonumber(val)
	return ( n ~= nil and n >= 0 )
end

function float(val)
	return ( tonumber(val) ~= nil )
end

function ipaddr(val)
	return ip4addr(val) or ip6addr(val)
end

function ip4addr(val)
	if val then
		return ip.IPv4(val) and true or false
	end

	return false
end

function ip4prefix(val)
	val = tonumber(val)
	return ( val and val >= 0 and val <= 32 )
end

function ip6addr(val)
	if val then
		return ip.IPv6(val) and true or false
	end

	return false
end

function ip6prefix(val)
	val = tonumber(val)
	return ( val and val >= 0 and val <= 128 )
end

function cidr4(val)
	local ip, mask = val:match("^([^/]+)/([^/]+)$")

	return ip4addr(ip) and ip4prefix(mask)
end

function cidr6(val)
	local ip, mask = val:match("^([^/]+)/([^/]+)$")

	return ip6addr(ip) and ip6prefix(mask)
end

function ipnet4(val)
	local ip, mask = val:match("^([^/]+)/([^/]+)$")

	return ip4addr(ip) and ip4addr(mask)
end

function ipnet6(val)
	local ip, mask = val:match("^([^/]+)/([^/]+)$")

	return ip6addr(ip) and ip6addr(mask)
end

function ipmask(val)
	return ipmask4(val) or ipmask6(val)
end

function ipmask4(val)
	return cidr4(val) or ipnet4(val) or ip4addr(val)
end

function ipmask6(val)
	return cidr6(val) or ipnet6(val) or ip6addr(val)
end

function ip6hostid(val)
	if val == "eui64" or val == "random" then
		return true
	else
		local addr = ip.IPv6(val)
		if addr and addr:prefix() == 128 and addr:lower("::1:0:0:0:0") then
			return true
		end
	end

	return false
end

function port(val)
	val = tonumber(val)
	return ( val and val >= 0 and val <= 65535 )
end

function portrange(val)
	local p1, p2 = val:match("^(%d+)%-(%d+)$")
	if p1 and p2 and port(p1) and port(p2) then
		return true
	else
		return port(val)
	end
end

function macaddr(val)
	return ip.checkmac(val) and true or false
end

function hostname(val, strict)
	if val and (#val < 254) and (
	   val:match("^[a-zA-Z_]+$") or
	   (val:match("^[a-zA-Z0-9_][a-zA-Z0-9_%-%.]*[a-zA-Z0-9]$") and
		val:match("[^0-9%.]")) or 
		val:match("^[htps]://[a-zA-Z0-9_][a-zA-Z0-9_%-%.]*[a-zA-Z0-9]$")
	) then
		return (not strict or not val:match("^_"))
	end
	return false
end

--Used for the hint
function netmask(val)
	return ip4addr(val)
end

function host(val, ipv4only)
	return hostname(val) or ((ipv4only == 1) and ip4addr(val)) or ((not (ipv4only == 1)) and ipaddr(val))
end

function network(val)
	return uciname(val) or host(val)
end

function hostport(val, ipv4only)
	local h, p = val:match("^([^:]+):([^:]+)$")
	return not not (h and p and host(h, ipv4only) and port(p))
end

function ip4addrport(val, bracket)
	local h, p = val:match("^([^:]+):([^:]+)$")
	return (h and p and ip4addr(h) and port(p))
end

function ipaddrport(val, bracket)
	local h, p = val:match("^([^%[%]:]+):([^:]+)$")
	if (h and p and ip4addr(h) and port(p)) then
		return true
	elseif (bracket == 1) then
		h, p = val:match("^%[(.+)%]:([^:]+)$")
		if  (h and p and ip6addr(h) and port(p)) then
			return true
		end
	end
	h, p = val:match("^([^%[%]]+):([^:]+)$")
	return (h and p and ip6addr(h) and port(p))
end

function wpakey(val)
	if #val == 64 then
		return (val:match("^[a-fA-F0-9]+$") ~= nil)
	else
		return (#val >= 8) and (#val <= 63)
	end
end

function wepkey(val)
	if val:sub(1, 2) == "s:" then
		val = val:sub(3)
	end

	if (#val == 10) or (#val == 26) then
		return (val:match("^[a-fA-F0-9]+$") ~= nil)
	else
		return (#val == 5) or (#val == 13)
	end
end

function hexstring(val)
	if val then
		return (val:match("^[a-fA-F0-9]+$") ~= nil)
	end
	return false
end

function hex(val, maxbytes)
	maxbytes = tonumber(maxbytes)
	if val and maxbytes ~= nil then
		return ((val:match("^0x[a-fA-F0-9]+$") ~= nil) and (#val <= 2 + maxbytes * 2))
	end
	return false
end

function base64(val)
	if val then
		return (val:match("^[a-zA-Z0-9/+]+=?=?$") ~= nil) and (math.fmod(#val, 4) == 0)
	end
	return false
end

function string(val)
	return true		-- Everything qualifies as valid string
end

function nospace(val)
	return (val:match("%s") == nil)
end

function directory(val, seen)
	local s = fs.stat(val)
	seen = seen or { }

	if s and not seen[s.ino] then
		seen[s.ino] = true
		if s.type == "dir" then
			return true
		elseif s.type == "lnk" then
			return directory( fs.readlink(val), seen )
		end
	end

	return false
end

function file(val, seen)
	local s = fs.stat(val)
	seen = seen or { }

	if s and not seen[s.ino] then
		seen[s.ino] = true
		if s.type == "reg" then
			return true
		elseif s.type == "lnk" then
			return file( fs.readlink(val), seen )
		end
	end

	return false
end

function device(val, seen)
	local s = fs.stat(val)
	seen = seen or { }

	if s and not seen[s.ino] then
		seen[s.ino] = true
		if s.type == "chr" or s.type == "blk" then
			return true
		elseif s.type == "lnk" then
			return device( fs.readlink(val), seen )
		end
	end

	return false
end

function uciname(val)
	return (val:match("^[a-zA-Z0-9_]+$") ~= nil)
end

function range(val, min, max)
	val = tonumber(val)
	min = tonumber(min)
	max = tonumber(max)

	if val ~= nil and min ~= nil and max ~= nil then
		return ((val >= min) and (val <= max))
	end

	return false
end

function min(val, min)
	val = tonumber(val)
	min = tonumber(min)

	if val ~= nil and min ~= nil then
		return (val >= min)
	end

	return false
end

function max(val, max)
	val = tonumber(val)
	max = tonumber(max)

	if val ~= nil and max ~= nil then
		return (val <= max)
	end

	return false
end

function rangelength(val, min, max)
	val = tostring(val)
	min = tonumber(min)
	max = tonumber(max)

	if val ~= nil and min ~= nil and max ~= nil then
		return ((#val >= min) and (#val <= max))
	end

	return false
end

function minlength(val, min)
	val = tostring(val)
	min = tonumber(min)

	if val ~= nil and min ~= nil then
		return (#val >= min)
	end

	return false
end

function maxlength(val, max)
	val = tostring(val)
	max = tonumber(max)

	if val ~= nil and max ~= nil then
		return (#val <= max)
	end

	return false
end

function phonedigit(val)
	return (val:match("^[0-9\*+#]+$") ~= nil)
end

function timehhmmss(val)
	local value = val:match("^[0-1][0-9]:[0-5][0-9]:[0-5][0-9]$")
	if value then
		return value
	end
	value = val:match("^[0-2][0-3]:[0-5][0-9]:[0-5][0-9]$")
	if value then
		return value
	end
	return false
end


function dateyyyymmdd(val)
	if val ~= nil then
		yearstr, monthstr, daystr = val:match("^(%d%d%d%d)-(%d%d)-(%d%d)$")
		if (yearstr == nil) or (monthstr == nil) or (daystr == nil) then
			return false;
		end
		year = tonumber(yearstr)
		month = tonumber(monthstr)
		day = tonumber(daystr)
		if (year == nil) or (month == nil) or (day == nil) then
			return false;
		end

		local days_in_month = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }

		local function is_leap_year(year)
			return (year % 4 == 0) and ((year % 100 ~= 0) or (year % 400 == 0))
		end

		function get_days_in_month(month, year)
			if (month == 2) and is_leap_year(year) then
				return 29
			else
				return days_in_month[month]
			end
		end
		if (year < 2015) then
			return false
		end
		if ((month == 0) or (month > 12)) then
			return false
		end
		if ((day == 0) or (day > get_days_in_month(month, year))) then
			return false
		end
		return true
	end
	return false
end

function rangelength(val, min, max, regex)
	val = tostring(val)
	min = tonumber(min)
	max = tonumber(max)

	if val ~= nil and min ~= nil and max ~= nil then
		if ((#val >= min) and (#val <= max)) then
			return (regex) and	(val:match(regex) ~= nil) or true
		end
	end

	return false
end

function fieldvalidation(val, valmat, minlen)
	val = tostring(val)
	valmat = tostring(valmat)
	minlen = tonumber(minlen)

	if #val >= minlen then
		return (val:match(valmat) ~= nil)
	end
	return false
end

function lengthvalidation(val, min, max, regex)
	if #val >= min and #val <= max then
		if regex then
			return (val:match(regex) ~= nil)
		else
			return true
		end
	end
	return false
end

function root_password(val)
	if val and (#val >= 8) and (#val <= 32) and
		val:match("[a-z]") and
		val:match("[A-Z]") and
		val:match("[0-9]") then
		return true
	end
	return false
end

function username_validate(val)
	return true
end

function password_validate(val)
	return true
end

function string_not_empty(val)
	return true
end

function geofencing(val, min, max)
	local string_val = val
	val = tonumber(val)
	min = tonumber(min)
	max = tonumber(max)

	if (float(val) and min and max) then
		local decimal_part, floating_part = string_val:match("([^.]*)%.([^.]*)")
		if (floating_part and floating_part:len() == 6) then
			if (val >= min and val <= max) then
				return true
			end
		end
	end
	return false
end

function protourl(value)
	local proto, host = value:match("(%w+)://([%w-.]+)")

	return (proto and host) and true or false
end

function url(value)
	local proto, url, path = value:match("(%w+)://([%w%.%-:]+)")
	if url == nil then 
		url, path = value:match("([%w%.%-:]+)/([%w%/%-%_]*)")
	end

	if url == nil then
		local proto, url, path = value:match("(%w+)://([%w%.%-:]+)/([%w%/%-%_]*)")
	end
    if url == nil then
		url = value:match("([%w%.%-:]*)")
		url = #url == #value and url or nil
	end
    return url ~= nil
end

function loglimit(value)
	if value and (value:match("^[0-9]+/second)") or
	value:match("^[0-9]+/minute)") or 
	value:match("^[0-9]+/hour)")) then
		return true
	end
	return false
end

function email(value)
	return value:match("^[a-zA-Z0-9._%%+-]+@[a-zA-Z0-9.-]+[.][a-zA-Z]+$")
end

function track(val, elemId, mode)
	return true
end

function time(val)
	if val and (val:match("^[0-1][0-9]:[0-5][0-9]$") or 
		val:match("^[0-9]:[0-5][0-9]$") or 
		val:match("^[0-2][0-3]:[0-5][0-9]$")) then
		return true
	end
	return false
end

function FF_or_0(value)
	if value and value == "0" or value == "65280" then
		return true
	else
		return false
	end
end